/*
 * StateFS oFono provider
 *
 * Properties translation from ofono values is performed in
 * contextkit-compatible way.
 *
 * Copyright (C) 2013, 2014 Jolla Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 */

#include <QTimer>

#include "provider_ofono.hpp"
#include "qdbusxml2cpp_dbus_types.hpp"
#include <qtaround/dbus.hpp>

#include <math.h>
#include <iostream>
#include <set>
#include <type_traits>

// until it appear in the cor library
template <typename T, typename FnT> void for_each
(FnT fn, typename std::enable_if<std::is_enum<T>::value>::type* = 0)
{
    static const auto begin = cor::enum_index(T::First_)
        , end = cor::enum_index(T::Last_) + 1;
    static_assert(end > begin, "Last_ >= First_");
    for (auto i = begin; i < end; ++i)
        fn(static_cast<T>(i));
}

#ifdef DEBUG
#define DBG qDebug
#else
struct null_stream {};
template <typename T>
null_stream operator << (null_stream dst, T)
{
    return dst;
}
#define DBG null_stream
#endif

namespace statefs { namespace ofono {

using statefs::qt::Namespace;
using statefs::qt::PropertiesSource;
using statefs::qt::make_proper_source;
using qtaround::dbus::sync;
using qtaround::dbus::async;

static char const *service_name = "org.ofono";

static const char *interface_names[] = {
    "AssistedSatelliteNavigation",
    "AudioSettings",
    "CallBarring",
    "CallForwarding",
    "CallMeter",
    "CallSettings",
    "CallVolume",
    "CellBroadcast",
    "ConnectionManager",
    "Handsfree",
    "LocationReporting",
    "MessageManager",
    "MessageWaiting",
    "NetworkRegistration",
    "Phonebook",
    "PushNotification",
    "RadioSettings",
    "SimManager",
    "SmartMessaging",
    "SimToolkit",
    "SupplementaryServices",
    "TextTelephony",
    "VoiceCallManager",
    "ExtSimInfo"
};

static_assert(sizeof(interface_names)/sizeof(interface_names[0])
              == (size_t)Interface::Last_ + 1, "Check interfaces list");

static QDebug operator << (QDebug dst, interfaces_set_type const &src)
{
    dst << "Cellular_interfaces=(";
    for_each<Interface>([&](Interface id) {
            auto i = cor::enum_index(id);
            if (src[i]) dst << interface_names[i] << ",";
        });
    dst << ")";
    return dst;
}


ModemManager::ModemManager(QDBusConnection &bus, QObject *parent)
    : QObject(parent)
    , watch_(bus, service_name)
    , bus_(bus)
{
}

void ModemManager::init()
{
    qDebug() << "Establish connection with ofono";

    auto process_modems = [this](PathPropertiesArray const &modems) {
        if (!manager_) {
            qDebug() << "Manager is reset, do not enumerate modems";
            return;
        }
        qDebug() << "There is(are) " << modems.size() << " modems";
        if (!modems.size())
            return;

        for (auto it = modems.begin(); it != modems.end(); ++it) {
            auto const &info = *it;
            auto path = std::get<0>(info).path();
            auto props = std::get<1>(info);
            emit modem_added(path, props);
        }
    };

    auto connect_manager = [this, process_modems]() {
        manager_.reset(new Manager(service_name, "/", bus_));
        using namespace std::placeholders;
        connect(manager_.get(), &Manager::ModemAdded
                , [this](QDBusObjectPath const &n, QVariantMap const&p) {
                    emit modem_added(n.path(), p);
                });
        connect(manager_.get(), &Manager::ModemRemoved
                , [this](QDBusObjectPath const &n) {
                    emit modem_removed(n.path());
                });

        async(this, manager_->GetModems(), process_modems);
    };

    auto reset_manager = [this]() {
        qDebug() << "Ofono is unregistered, cleaning up";
        emit reset();
    };
    watch_.init(connect_manager, reset_manager);
    connect_manager();
}



#define MK_IFACE_ID(name) {#name, Interface::name}

static const std::map<QString, Interface> interface_ids = {
    MK_IFACE_ID(AssistedSatelliteNavigation),
    MK_IFACE_ID(AudioSettings),
    MK_IFACE_ID(CallBarring),
    MK_IFACE_ID(CallForwarding),
    MK_IFACE_ID(CallMeter),
    MK_IFACE_ID(CallSettings),
    MK_IFACE_ID(CallVolume),
    MK_IFACE_ID(CellBroadcast),
    MK_IFACE_ID(ConnectionManager),
    MK_IFACE_ID(Handsfree),
    MK_IFACE_ID(LocationReporting),
    MK_IFACE_ID(MessageManager),
    MK_IFACE_ID(MessageWaiting),
    MK_IFACE_ID(NetworkRegistration),
    MK_IFACE_ID(Phonebook),
    MK_IFACE_ID(PushNotification),
    MK_IFACE_ID(RadioSettings),
    MK_IFACE_ID(SimManager),
    MK_IFACE_ID(SmartMessaging),
    MK_IFACE_ID(SimToolkit),
    MK_IFACE_ID(SupplementaryServices),
    MK_IFACE_ID(TextTelephony),
    MK_IFACE_ID(VoiceCallManager),
    MK_IFACE_ID(ExtSimInfo)
};

static interfaces_set_type get_interfaces(QStringList const &from)
{
    static const QString std_prefix = "org.ofono.";
    static const auto std_prefix_len = std_prefix.length();

    static const QString ext_prefix = "org.nemomobile.ofono.";
    static const auto ext_prefix_len = ext_prefix.length();

    interfaces_set_type res;
    for (auto const &v : from) {
        if (v.leftRef(ext_prefix_len) == ext_prefix) {
            auto p = interface_ids.find(v.mid(ext_prefix_len).prepend(QStringLiteral("Ext")));
            if (p != interface_ids.end())
                res.set((size_t)p->second);
        } else if (v.leftRef(std_prefix_len) == std_prefix) {
            auto p = interface_ids.find(v.mid(std_prefix_len));
            if (p != interface_ids.end())
                res.set((size_t)p->second);
        }
    }
    return res;
}

/**
 * @addtogroup statefs_properties
 *
 * @section cellular_ns Cellular namespace properties:
 *
 * - SignalStrength (integer, [0, 100]) - signal strength in %
 *
 * - DataTechnology (string, legacy, [gprs, egprs, hspa, umts, lte]) -
 *   cellular data technology
 *
 * - RegistrationStatus (string, legacy, [no-sim, offline, home,
 *   forbidden, roam]) - cellular network registration status.
 *
 * - Sim (string, [present, absent, ]) - SIM cqard presence if known
 *
 * - Status (string, [disabled, unregistered, registered, searching, denied,
 *   unknown, roaming]) - cellular registration status compatible with
 *   ofono values set, with the exception of "disabled" which indicates that
 *   the modem interface is not present.
 *
 * - Technology (string, [gsm, umts, lte]) - cellular technology
 *
 * - SignalBars (integer, legacy5]) - number of bars to show
 *   for signal strength
 *
 * - CellName (string)
 *
 * - NetworkName (string)
 *
 * - ExtendedNetworkName (string, legacy)
 *
 * - SubscriberIdentity (string)
 *
 * - CurrentMCC (integer) - current Mobile Country Code
 *
 * - CurrentMNC (integer) - current Mobile Network Code
 *
 * - HomeMCC (integer) - home Mobile Country Code
 *
 * - HomeMNC (integer) - home Mobile Network Code
 *
 * - StkIdleModeText (string)
 *
 * - MMSContext (string)
 *
 * - DataRoamingAllowed (boolean, [0, 1])
 *
 * - GPRSAttached (boolean, [0, 1])
 *
 * - CapabilityVoice (boolean, [0, 1]) - is voice calling available
 *
 * - CapabilityData (boolean, [0, 1]) - is mobile data transfer available
 *
 * - CallCount (integer)
 *
 * - ModemPath (string)
 *
 * - ServiceProviderName (string)
 *
 * - CachedCardIdentifier (string)
 *
 * - CachedSubscriberIdentity (string)
 */
static constexpr const char *property_names[] = {
    "SignalStrength",
    "DataTechnology",
    "RegistrationStatus",
    "Sim",
    "Status",
    "Technology",
    "SignalBars",
    "CellName",
    "NetworkName",
    "ExtendedNetworkName",
    "SubscriberIdentity",
    "CurrentMCC",
    "CurrentMNC",
    "HomeMCC",
    "HomeMNC",
    "StkIdleModeText",
    "MMSContext",
    "DataRoamingAllowed",
    "GPRSAttached",
    "CapabilityVoice",
    "CapabilityData",
    "CallCount",
    "ModemPath",
    "ServiceProviderName",
    "CachedCardIdentifier",
    "CachedSubscriberIdentity"
};

static_assert(sizeof(property_names)/sizeof(property_names[0])
              == (size_t)Property::Last_ + 1, "Check properties list");

constexpr char const *Bridge::name(Property id)
{
    return property_names[cor::enum_index(id)];
}

static char const * sim_presence_name(SimPresent v)
{
    switch(v) {
    case SimPresent::Yes:
        return "present";
        break;
    case SimPresent::No:
        return "absent";
        break;
    default:
        return "";
    }
}

enum class State { UnchangedReset = 0, UnchangedSet = 1, Reset = 2, Set = (1 | 2) };

static inline bool is_set(State s)
{
    return (static_cast<int>(s) & 1);
}

static inline bool is_changed(State s)
{
    return (static_cast<int>(s) & 2);
}

template <typename T>
static bool is_set(std::bitset<cor::enum_size<T>()> const &src, T id)
{
    return src[cor::enum_index(id)];
}

static State get_state_change(interfaces_set_type const &before
                              , interfaces_set_type const &now
                              , Interface id)
{
    auto i = (size_t)id;
    auto from = before[i], to = now[i];
    return (from == to
            ? (to ? State::UnchangedSet : State::UnchangedReset)
            : (to ? State::Set : State::Reset));
}

template <typename K, typename F, typename ... Args>
void map_exec(std::map<K, F> const &fns, K const &k, Args&&... args)
{
    auto pfn = fns.find(k);
    if (pfn != fns.end())
        pfn->second(std::forward<Args>(args)...);
}

template <typename K, typename F, typename AltFn, typename ... Args>
void map_exec_or(std::map<K, F> const &fns, AltFn alt_fn, K const &k, Args&&... args)
{
    auto pfn = fns.find(k);
    if (pfn != fns.end())
        pfn->second(std::forward<Args>(args)...);
    else
        alt_fn(k, std::forward<Args>(args)...);
}

template <typename T, typename K, typename F, typename ... Args>
void map_member_exec(T *self, std::map<K, F> const &fns, K const &k, Args&&... args)
{
    auto pfn = fns.find(k);
    if (pfn != fns.end()) {
        auto fn = pfn->second;
        (self->*fn)(std::forward<Args>(args)...);
    }
}

template <typename FnT>
bool find_process_object(PathPropertiesArray const &src, FnT fn)
{
    for (auto it = src.begin(); it != src.end(); ++it) {
        auto const &info = *it;
        auto path = std::get<0>(info).path();
        auto props = std::get<1>(info);

        if (fn(path, props))
            return true;
    }
    return false;
}

typedef Bridge::Status Status;


QDebug & operator << (QDebug &dst, Status src)
{
    static const char *names[] = {
        "Disabled", "Offline", "Registered", "Searching"
        , "Denied", "Unknown", "Roaming"
    };
    static_assert(sizeof(names)/sizeof(names[0]) == cor::enum_size<Status>()
                  , "Check Status values names");
    dst << names[size_t(src)];
    return dst;
}

typedef std::map<QString, std::pair<char const *, char const *> > tech_map_type;
typedef std::map<QString, Status> status_map_type;
typedef std::array<QString, cor::enum_size<Status>()> status_array_type;


static const tech_map_type tech_map_ = {
    {"gsm", {"gsm", "gprs"}}
    , {"edge", {"gsm", "egprs"}}
    , {"hspa", {"umts", "hspa"}}
    , {"umts", {"umts", "umts"}}
    , {"lte", {"lte", "lte"}}
};

static const status_map_type status_map_ = {
    {"", Status::Disabled}
    , {"unregistered", Status::Offline}
    , {"registered", Status::Registered}
    , {"searching", Status::Searching}
    , {"denied", Status::Denied}
    , {"unknown", Status::Unknown}
    , {"roaming", Status::Roaming}
};

Status Bridge::map_status(QString const &name)
{
    auto it = status_map_.find(name);
    return (it != status_map_.end()) ? it->second : Status::Offline;
}

QString const & Bridge::ckit_status(Status status, SimPresent sim)
{
    static const QString names[] = {
        "disabled", "offline", "home"
        , "offline", "forbidden", "offline", "roam"
    };
    static const QString no_sim("no-sim");

    static_assert(sizeof(names)/sizeof(names[0]) == cor::enum_size<Status>()
                  , "Check Status values names");

    return (sim == SimPresent::Yes
            ? names[static_cast<size_t>(status)]
            : no_sim);
}

QString const & Bridge::ofono_status(Status status)
{
    static const QString names[] = {
            "disabled", "unregistered", "registered"
            , "searching", "denied", "unknown", "roaming"
        };
    static_assert(sizeof(names)/sizeof(names[0]) == cor::enum_size<Status>()
                  , "Check Status values names");

    return names[static_cast<size_t>(status)];
}

// read in big endian
static const std::bitset<cor::enum_size<Status>()> status_registered_("1000100");

static const std::map<QString, Property> sim_props_map_ = {
    { "MobileCountryCode", Property::HomeMCC }
    , { "MobileNetworkCode", Property::HomeMNC }
    , { "SubscriberIdentity", Property::SubscriberIdentity }
};

static const std::map<QString, Property> stk_props_map_ = {
    { "IdleModeText", Property::StkIdleModeText }
};

static Bridge::property_action_type direct_update(Property id)
{
    return [id](Bridge *self, QVariant const &v) {
        self->updateProperty(id, v);
    };
}

static Bridge::property_action_type bind_member
(void (Bridge::*fn)(QVariant const&))
{
    using namespace std::placeholders;
    return [fn](Bridge *self, QVariant const &v) {
        (self->*fn)(v);
    };
}

const Bridge::property_map_type Bridge::net_property_actions_ = {
    { "Name", bind_member(&Bridge::set_network_name)}
    , { "Strength", [](Bridge *self, QVariant const &v) {
            auto strength = v.toUInt();
            self->updateProperty(Property::SignalStrength, strength);
            // 0-5
            self->updateProperty(Property::SignalBars, (strength + 19) / 20);
        } }
    , { "Status", [](Bridge *self, QVariant const &v) {
            qDebug() << "Ofono status " << v.toString();
            self->updateProperty(Property::Status, v);
            self->set_status(self->map_status(v.toString()));
        } }
    , { "MobileCountryCode", direct_update(Property::CurrentMCC) }
    , { "MobileNetworkCode", direct_update(Property::CurrentMNC) }
    , { "CellId", direct_update(Property::CellName) }
    , { "Technology", [](Bridge *self, QVariant const &v) {
            auto tech = v.toString();
            auto pt = tech_map_.find(tech);
            if (pt != tech_map_.end()) {
                auto tech_dtech = pt->second;
                self->updateProperty(Property::Technology, tech_dtech.first);
                self->updateProperty(Property::DataTechnology, tech_dtech.second);
            }
        } }
};

const Bridge::property_map_type Bridge::operator_property_actions_ = {
    { "Name", bind_member(&Bridge::set_operator_name)}
};

const Bridge::property_map_type Bridge::connman_property_actions_ = {
    { "RoamingAllowed", direct_update(Property::DataRoamingAllowed) }
    , { "Attached", direct_update(Property::GPRSAttached) }
};

static const std::map<QString, Property> ext_sim_props_map_ = {
    { "spn", Property::ServiceProviderName }
    , { "iccid", Property::CachedCardIdentifier }
    , { "imsi", Property::CachedSubscriberIdentity }
};

int Bridge::active_bridges_ = 0;

Bridge::Bridge(MainNs *ns, QDBusConnection &bus, ModemManager *manager, const QRegularExpression &modem_pattern)
    : PropertiesSource(ns)
    , bus_(bus)
    , sim_present_(SimPresent::Unknown)
    , status_(Status::Unknown)
    , network_name_{"", ""}
    , set_name_(&Bridge::set_name_home)
    , modem_pattern_(modem_pattern)
{
    connect(manager, &ModemManager::modem_added, this, &Bridge::setup_modem);
    connect(manager, &ModemManager::modem_removed, this, &Bridge::reset_modem);
    connect(manager, &ModemManager::reset, this, &Bridge::reset);
}

void Bridge::reset()
{
    QTimer::singleShot(0, this, SLOT(update_capabilities()));

    if (!modem_path_.isEmpty()) {
        active_bridges_--;
        modem_path_.clear();
    }
    network_.reset();
    operator_.reset();
    stk_.reset();
    sim_.reset();
    sim_present_ = SimPresent::Unknown;
    callManager_.reset();
    calls_.clear();
    modem_.reset();
    interfaces_.reset();
    simInfo_.reset();
    reset_props();
}

void Bridge::set_network_name(QVariant const &v)
{
    network_name_.first = v.toString();
    (this->*set_name_)();
}

void Bridge::set_operator_name(QVariant const &v)
{
    network_name_.second = v.toString();
    (this->*set_name_)();
}

void Bridge::set_name_home()
{
    auto name = network_name_.first;
    if (!name.size())
        name = network_name_.second;
    updateProperty(Property::NetworkName, name);
    updateProperty(Property::ExtendedNetworkName, name);
}

void Bridge::set_name_roaming()
{
    auto name = network_name_.first;
    if (!name.size())
        name = network_name_.second;
    updateProperty(Property::NetworkName, name);
    updateProperty(Property::ExtendedNetworkName, name);
}

void Bridge::set_status(Status new_status)
{
    if (new_status == status_)
        return;

    qDebug() << "Cellular status" << status_ << "->" << new_status;

    auto expected = (new_status == Status::Roaming
                     ? &Bridge::set_name_roaming
                     : &Bridge::set_name_home);

    if (expected != set_name_)
        set_name_ = expected;

    auto is_registered = is_set(status_registered_, new_status);
    auto was_registered = is_set(status_registered_, status_);
    auto is_changed = (was_registered != is_registered);

    updateProperty(Property::RegistrationStatus, ckit_status(new_status, sim_present_));

    status_ = new_status;
    if (is_changed) {
        qDebug() << (is_registered ? "Registered" : "Unregistered");
        if (is_registered) {
            if (!modem_) {
                qWarning() << "Network w/o modem?";
            } else if (!sim_) {
                setup_sim(modem_path_);
            } else if (!network_) {
                setup_network(modem_path_);
            } else {
                enumerate_operators();
            }
        }
    }
}

void Bridge::update_mms_context()
{
    mmsContext_ = QString();
    for (auto iter = connectionContexts_.begin(); iter != connectionContexts_.end(); ++iter) {
        if (iter->second.properties["Type"].toString() == "mms"
                && !iter->second.properties["MessageCenter"].toString().isEmpty()) {
            mmsContext_ = iter->first;
            break;
        }
    }
    updateProperty(Property::MMSContext, mmsContext_);
    DBG() << "updated MMS context" << mmsContext_;
}

void Bridge::reset_props()
{
    static const auto status = Status::Disabled;
    set_status(status);
    static_cast<MainNs*>(target_)->resetProperties(status, sim_present_, active_bridges_ > 0);
}

void Bridge::reset_modem(const QString &path)
{
    QTimer::singleShot(0, this, SLOT(update_capabilities()));

    if (path.isEmpty() || path == modem_path_) {
        qDebug() << "Reset modem properties" << path;
        active_bridges_--;
        modem_path_ = "";
        modem_.reset();
        reset_connectionManager();
        reset_props();
    }
}

void Bridge::reset_sim()
{
    qDebug() << "Reset sim properties";
    sim_present_ = SimPresent::Unknown;
    sim_.reset();
    interfaces_.reset((size_t)Interface::SimManager);
    reset_props();
}

void Bridge::reset_callManager()
{
    qDebug() << "Reset VoiceCallManager properties";
    callManager_.reset();
    calls_.clear();
    interfaces_.reset((size_t)Interface::VoiceCallManager);
    updateProperty(Property::CallCount, 0);
}

void Bridge::reset_simInfo()
{
    qDebug() << "Reset ExtSimInfo properties";
    simInfo_.reset();
    interfaces_.reset((size_t)Interface::ExtSimInfo);
    updateProperty(Property::ServiceProviderName , "");
    updateProperty(Property::CachedCardIdentifier , "");
    updateProperty(Property::CachedSubscriberIdentity , "");
}

void Bridge::reset_network()
{
    qDebug() << "Reset cellular network properties";
    operator_.reset();
    network_.reset();
    interfaces_.reset((size_t)Interface::NetworkRegistration);
    reset_props();
}

void Bridge::reset_stk()
{
    qDebug() << "Reset sim toolkit properties";
    stk_.reset();
    interfaces_.reset((size_t)Interface::SimToolkit);
    updateProperty(Property::StkIdleModeText, "");
}

void Bridge::process_interfaces(QStringList const &v)
{
    auto interfaces = get_interfaces(v);
    qDebug() << interfaces;

    auto on_exit = cor::on_scope_exit([this, interfaces]() {
            interfaces_ = interfaces;
        });

    auto state = [this, interfaces](Interface id) {
        return get_state_change(interfaces_, interfaces, id);
    };

    auto sim_state = state(Interface::SimManager);
    if (sim_state == State::Reset) {
        if (sim_) {
            qDebug() << "SimManager is gone";
            reset_sim();
            return;
        }
    } else if (sim_state == State::Set) {
        setup_sim(modem_path_);
    }

    auto net_state = state(Interface::NetworkRegistration);
    if (net_state == State::Reset) {
        reset_network();
    } else if (net_state == State::Set) {
        if (is_set(sim_state))
            qWarning() << "Cellular NetworkRegistration w/o SimManager!";
        setup_network(modem_path_);
    }

    auto stk_state = state(Interface::SimToolkit);
    if (stk_state == State::Set)
        setup_stk(modem_path_);
    else if (stk_state == State::Reset)
        reset_stk();

    auto cm_state = state(Interface::ConnectionManager);
    if (cm_state == State::Set)
        setup_connectionManager(modem_path_);
    else if (cm_state == State::Reset)
        reset_connectionManager();

    auto calls_state = state(Interface::VoiceCallManager);
    if (calls_state == State::Set)
        setup_callManager(modem_path_);
    else if (calls_state == State::Reset)
        reset_callManager();

    auto simInfo_state = state(Interface::ExtSimInfo);
    if (simInfo_state == State::Set)
        setup_simInfo(modem_path_);
    else if (simInfo_state == State::Reset)
        reset_simInfo();
}

bool Bridge::setup_modem(QString const &path, QVariantMap const &props)
{
    QTimer::singleShot(0, this, SLOT(update_capabilities()));

    if (!modem_path_.isEmpty()) {
        // Already handling a different modem.
        return false;
    }

    if (props["Type"].toString() != "hardware") {
        // TODO hardcoded for phones now, no support for e.g. DUN
        return false;
    }

    QRegularExpressionMatch match = modem_pattern_.match(path);
    if (!match.hasMatch()) {
        // This modem doesn't match the pattern we want
        return false;
    }

    qDebug() << "Hardware modem " << path << modem_pattern_.pattern();

    auto update = [this](QString const &n, QVariant const &v) {
        DBG() << "Modem prop: " << n << "=" << v;
        if (n == "Interfaces")
            process_interfaces(v.toStringList());
        else if (n == "Powered") {
            auto is_powered = v.toBool();
            qDebug() << "Modem power is" << (is_powered ? "on" : "off");
            if (!is_powered)
                reset_props();
        }
    };

    modem_.reset(new Modem(service_name, path, bus_));
    modem_path_ = path;
    active_bridges_++;

    connect(modem_.get(), &Modem::PropertyChanged
            , [update](QString const &n, QDBusVariant const &v) {
                update(n, v.variant());
            });
    auto process_properties = [this, update](QVariantMap const &properties) {
        for (auto it = properties.begin(); it != properties.end(); ++it)
            update(it.key(), it.value());
    };
    async(this, modem_->GetProperties(), process_properties);

    updateProperty(Property::ModemPath, path);

    return true;
}

void Bridge::set_sim_presence(SimPresent v)
{
    if (v == sim_present_)
        return;

    sim_present_ = v;
    updateProperty(Property::Sim, sim_presence_name(v));
    if (sim_present_ == SimPresent::No) {
        qDebug() << "Ofono: no sim";
        set_status(Status::Offline);
    } else if (sim_present_ == SimPresent::Yes) {
        qDebug() << "Ofono: sim is present";
        if (!network_)
            set_status(Status::Offline);
        else if (!modem_path_.isEmpty())
            setup_network(modem_path_);
    }
}

void Bridge::update_capabilities()
{
    // data and voice capabilities become true if there's any modem tracked.
    // as of now assuming that all modems support both features.
    updateProperty(Property::CapabilityData, active_bridges_ > 0);
    updateProperty(Property::CapabilityVoice, active_bridges_ > 0);
}

bool Bridge::setup_operator(QString const &path, QVariantMap const &props)
{
    qDebug() << "Operator " << props["Name"];
    auto status = props["Status"].toString();
    if (status != "current")
        return false;

    qDebug() << "Setup current operator properties";

    auto update = [this](QString const &n, QVariant const &v) {
        DBG() << "Operator prop: " << n << "=" << v;
        if (sim_present_ == SimPresent::No)
            qDebug() << "No sim, operator property" << n << "->" << v;

        map_exec(operator_property_actions_, n, this, v);
    };

    operator_.reset(new Operator(service_name, path, bus_));
    operator_path_ = path;
    connect(operator_.get(), &Operator::PropertyChanged
            , [update](QString const &n, QDBusVariant const &v) {
                update(n, v.variant());
            });
    for (auto it = props.begin(); it != props.end(); ++it)
        update(it.key(), it.value());
    return true;
}

void Bridge::init()
{
}

void Bridge::setup_network(QString const &path)
{
    qDebug() << "Get cellular network properties";
    auto update = [this](QString const &n, QVariant const &v) {
        DBG() << "Network: prop" << n << "=" << v;
        if (sim_present_ == SimPresent::No)
            qDebug() << "No sim, network prop" << n << "->" << v;

        map_exec(net_property_actions_, n, this, v);
    };

    network_.reset(new Network(service_name, path, bus_));

    DBG() << "Connect Network::PropertyChanged";
    connect(network_.get(), &Network::PropertyChanged
            , [update](QString const &n, QDBusVariant const &v) {
                update(n, v.variant());
            });

    auto process_props = [this, update](QVariantMap const &props) {
        if (!network_) {
            qDebug() << "Network is reset, do not process props";
            return;
        }
        for (auto it = props.begin(); it != props.end(); ++it)
            update(it.key(), it.value());
        enumerate_operators();
    };
    async(this, network_->GetProperties(), process_props);
}

void Bridge::setup_stk(QString const &path)
{
    qDebug() << "Get SimToolkit properties";
    auto update = [this](QString const &n, QVariant const &v) {
        DBG() << "SimToolkit: prop" << n << "=" << v;
        if (sim_present_ == SimPresent::No)
            DBG() << "No sim, SimToolkit prop" << n << "->" << v;
        auto it = stk_props_map_.find(n);
        if (it != stk_props_map_.end())
            updateProperty(it->second, v);
    };

    stk_.reset(new SimToolkit(service_name, path, bus_));
    connect(stk_.get(), &SimToolkit::PropertyChanged
            , [update](QString const &n, QDBusVariant const &v) {
                update(n, v.variant());
            });
    auto res = sync(stk_->GetProperties());
    if (res.isError()) {
        qWarning() << "SimToolkit GetProperties error:" << res.error();
        return;
    }

    auto props = res.value();
    for (auto it = props.begin(); it != props.end(); ++it)
        update(it.key(), it.value());
}

void Bridge::reset_connectionManager()
{
    qDebug() << "Reset connection manager";
    connectionManager_.reset();
    interfaces_.reset((size_t)Interface::ConnectionManager);
    connectionContexts_.clear();
    update_mms_context();
}

void Bridge::setup_connectionManager(QString const &path)
{
    qDebug() << "Setup connection manager" << path;

    auto update = [this](QString const &n, QVariant const &v) {
        DBG() << "CM prop: " << n << "=" << v;
        map_exec(connman_property_actions_, n, this, v);
    };

    auto contextAdded = [this](QDBusObjectPath const &c, QVariantMap const &m) {
        DBG() << "CM: context added" << c.path() << "=" << m;

        const QString &contextPath = c.path();
        ConnectionCache &connection = connectionContexts_[contextPath];
        connection.context.reset(new ConnectionContext(service_name, contextPath, bus_));

        connect(connection.context.get(), &ConnectionContext::PropertyChanged
                , [this,contextPath](QString const &p, QDBusVariant const &v) {
                    connectionContexts_[contextPath].properties.insert(p, v.variant());
                    update_mms_context();
                });

        connection.properties = m;
        update_mms_context();
    };

    auto contextRemoved = [this](QDBusObjectPath const &c) {
        DBG() << "CM: context removed" << c.path();
        connectionContexts_.erase(c.path());
        update_mms_context();
    };

    connectionManager_.reset(new ConnectionManager(service_name, path, bus_));

    connect(connectionManager_.get(), &ConnectionManager::PropertyChanged
            , [update](QString const &n, QDBusVariant const &v) {
                update(n, v.variant());
            });
    connect(connectionManager_.get(), &ConnectionManager::ContextAdded
            , [contextAdded](QDBusObjectPath const &c, QVariantMap const &m) {
                contextAdded(c, m);
            });
    connect(connectionManager_.get(), &ConnectionManager::ContextRemoved
            , [contextRemoved](QDBusObjectPath const &c) {
                contextRemoved(c);
            });

    auto res = sync(connectionManager_->GetProperties());
    if (res.isError()) {
        qWarning() << "ConnectionManager GetProperties error:" << res.error();
        return;
    }
    auto props = res.value();
    for (auto it = props.begin(); it != props.end(); ++it)
        update(it.key(), it.value());

    auto contexts_res = sync(connectionManager_->GetContexts());
    if (contexts_res.isError()) {
        qWarning() << "ConnectionManager GetContexts error:" << contexts_res.error();
        return;
    }

    auto contexts = contexts_res.value();
    DBG() << "Got contexts" << contexts.count();
    for (auto it = contexts.begin(); it != contexts.end(); ++it) {
        auto const &info = *it;
        contextAdded(std::get<0>(info), std::get<1>(info));
    }
}

void Bridge::enumerate_operators()
{
    if (!network_) {
        qWarning() << "Can't enumerate operators, network is null";
        return;
    }
    auto process_operators = [this](PathPropertiesArray const &ops) {
        if (!network_) {
            qDebug() << "network is null, skip operators";
            return;
        }
        using namespace std::placeholders;
        auto process = std::bind(&Bridge::setup_operator, this, _1, _2);
        find_process_object(ops, process);
    };
    async(this, network_->GetOperators(), process_operators);
}

void Bridge::setup_sim(QString const &path)
{
    qDebug() << "Get sim properties";
    auto update = [this](QString const &n, QVariant const &v) {
        DBG() << "Sim prop: " << n << "=" << v;
        if (n == "Present") {
            set_sim_presence(v.toBool() ? SimPresent::Yes : SimPresent::No);
        } else {
            auto it = sim_props_map_.find(n);
            if (it != sim_props_map_.end())
                updateProperty(it->second, v);
        }
    };
    sim_.reset(new SimManager(service_name, path, bus_));
    connect(sim_.get(), &SimManager::PropertyChanged
            , [update](QString const &n, QDBusVariant const &v) {
                update(n, v.variant());
            });

    auto res = sync(sim_->GetProperties());
    if (res.isError()) {
        qWarning() << "Sim GetProperties error:" << res.error();
        return;
    }
    auto props = res.value();
    for (auto it = props.begin(); it != props.end(); ++it)
        update(it.key(), it.value());

    if (is_set(interfaces_, Interface::SimToolkit))
        setup_stk(modem_path_);
}

void Bridge::setup_callManager(QString const &path)
{
    qDebug() << "Setup VoiceCallManager";

    auto process_calls = [this](PathPropertiesArray const &calls) {
        for (auto it = calls.begin(); it != calls.end(); ++it)
            calls_.insert(std::get<0>(*it).path());
        updateProperty(Property::CallCount, QVariant::fromValue(calls_.size()));
    };

    callManager_.reset(new VoiceCallManager(service_name, path, bus_));

    connect(callManager_.get(), &VoiceCallManager::CallAdded
            , [this](QDBusObjectPath const &n, QVariantMap const &) {
                    calls_.insert(n.path());
                    updateProperty(Property::CallCount,  QVariant::fromValue(calls_.size()));
            });

    connect(callManager_.get(), &VoiceCallManager::CallRemoved
            , [this](QDBusObjectPath const &n) {
                    calls_.erase(n.path());
                    updateProperty(Property::CallCount,  QVariant::fromValue(calls_.size()));
            });

    async(this, callManager_->GetCalls(), process_calls);
}

void Bridge::setup_simInfo(QString const &path)
{
    qDebug() << "Setup ExtSimInfo";

    auto update = [this](QString const &n, QVariant const &v) {
        DBG() << "ExtSim prop: " << n << "=" << v;
        auto it = ext_sim_props_map_.find(n);
        if (it != ext_sim_props_map_.end())
            updateProperty(it->second, v);
    };

    simInfo_.reset(new ExtSimInfo(service_name, path, bus_));
    connect(simInfo_.get(), &ExtSimInfo::ServiceProviderNameChanged
            , [update](QString const &spn) {
                update("spn", spn);
            });
    connect(simInfo_.get(), &ExtSimInfo::CardIdentifierChanged
            , [update](QString const &iccid) {
                update("iccid", iccid);
            });
    connect(simInfo_.get(), &ExtSimInfo::SubscriberIdentityChanged
            , [update](QString const &imsi) {
                update("imsi", imsi);
            });

    async(this, simInfo_->GetServiceProviderName(), [update](QString const &spn) {
        update("spn", spn);
    });
    async(this, simInfo_->GetCardIdentifier(), [update](QString const &iccid) {
        update("iccid", iccid);
    });
    async(this, simInfo_->GetSubscriberIdentity(), [update](QString const &imsi) {
        update("imsi", imsi);
    });
}

void MainNs::resetProperties(Bridge::Status status, SimPresent sim, bool voice_and_data)
{
    qDebug() << "Reset properties";
    setProperties(defaults_);
    updateProperty(Property::RegistrationStatus, Bridge::ckit_status(status, sim));
    updateProperty(Property::Sim, sim_presence_name(sim));
    updateProperty(Property::CapabilityVoice, voice_and_data);
    updateProperty(Property::CapabilityData, voice_and_data);
}

// TODO 2 contexkit properties are not supported yet:
// Phone.Call and Phone.Muted
// There is no components using it so the question
// is should they be supported at all

#define PROP_(prop_name, value) { Bridge::name(Property::prop_name), value }

MainNs::MainNs(QDBusConnection &bus, const char *name, ModemManager *manager, const QRegularExpression &modem_pattern, statefs_provider_mode mode)
    : Namespace(name, make_proper_source<Bridge>(mode, this, bus, manager, modem_pattern))
    , defaults_({
            PROP_(SignalStrength, "0")
                , PROP_(DataTechnology, "unknown")
                , PROP_(Status, "disabled")
                , PROP_(Technology, "unknown")
                , PROP_(SignalBars, "0")
                , PROP_(CellName, "")
                , PROP_(NetworkName, "")
                , PROP_(ExtendedNetworkName, "")
                , PROP_(SubscriberIdentity, "")
                , PROP_(CurrentMCC, "0")
                , PROP_(CurrentMNC, "0")
                , PROP_(HomeMCC, "0")
                , PROP_(HomeMNC, "0")
                , PROP_(StkIdleModeText, "")
                , PROP_(MMSContext, "")
                , PROP_(DataRoamingAllowed, "0")
                , PROP_(GPRSAttached, "0")
                , PROP_(CapabilityVoice, "0")
                , PROP_(CapabilityData, "0")
                , PROP_(CallCount, "0")
                , PROP_(ModemPath, "")
                , PROP_(ServiceProviderName, "")
                , PROP_(CachedCardIdentifier, "")
                , PROP_(CachedSubscriberIdentity, "")
        })
{
    // contextkit prop
    static auto const sim = SimPresent::Unknown;
    auto status = Bridge::ckit_status(Status::Offline, sim);
    addProperty(Bridge::name(Property::RegistrationStatus), status.toUtf8());
    addProperty(Bridge::name(Property::Sim), sim_presence_name(sim));

    for (auto v : defaults_)
        addProperty(v.first, v.second);

    src_->init();
}

#undef PROP_

class Provider;
static Provider *provider = nullptr;

class Provider : public statefs::AProvider
{
public:
    Provider(statefs_server *server)
        : AProvider("ofono", server)
        , bus_(QDBusConnection::systemBus())
    {
        if (!server || server->mode == statefs_provider_mode_run)
            manager_.reset(new ModemManager(bus_));

        // We need the modem order to be static, so we look for modems with _0 and _1, or phonesim.
        // Perhaps needs a config file, or perhaps we'll get a new ofono API. This will do for now.
        auto ns = std::make_shared<MainNs>
            (bus_, "Cellular", manager_.get(), QRegularExpression(QStringLiteral(".*(_0|phonesim)$"))
             , server ? server->mode : statefs_provider_mode_run);
        insert(std::static_pointer_cast<statefs::ANode>(ns));

        ns = std::make_shared<MainNs>
            (bus_, "Cellular_1", manager_.get(), QRegularExpression(QStringLiteral(".*(_1|phonesim)$"))
             , server ? server->mode : statefs_provider_mode_run);
        insert(std::static_pointer_cast<statefs::ANode>(ns));

        if (manager_)
            manager_->init();
    }
    virtual ~Provider() {}

    virtual void release() {
        if (this == provider) {
            delete provider;
            provider = nullptr;
        }
    }

private:
    QDBusConnection bus_;
    std::unique_ptr<ModemManager> manager_;
};

static inline Provider *init_provider(statefs_server *server)
{
    registerDataTypes();
    if (provider)
        throw std::logic_error("provider ptr is already set");
    provider = new Provider(server);
    return provider;
}

}}

EXTERN_C struct statefs_provider * statefs_provider_get
(struct statefs_server *server)
{
    return statefs::ofono::init_provider(server);
}
